# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Modifications copyright (C) 2024 S.Cao
# ported Google's Jax-CFD functional template to PyTorch's tensor ops

"""Functions for computing and applying pressure."""

from typing import Callable, Optional

import torch

from . import grids, fast_diagonalization as solver, finite_differences as fd


Array = grids.Array
GridArray = grids.GridArray
GridArrayVector = grids.GridArrayVector
GridVariable = grids.GridVariable
GridVariableVector = grids.GridVariableVector
BoundaryConditions = grids.BoundaryConditions


def _rhs_transform(
    u: GridArray,
    bc: BoundaryConditions,
) -> Array:
    """Transform the RHS of pressure projection equation for stability.

    In case of poisson equation, the kernel is subtracted from RHS for stability.

    Args:
      u: a GridArray that solves ∇²x = u.
      bc: specifies boundary of x.

    Returns:
      u' s.t. u = u' + kernel of the laplacian.
    """
    u_data = u.data
    for axis in range(u.grid.ndim):
        if (
            bc.types[axis][0] == grids.BCType.NEUMANN
            and bc.types[axis][1] == grids.BCType.NEUMANN
        ):
            # if all sides are neumann, poisson solution has a kernel of constant
            # functions. We substact the mean to ensure consistency.
            u_data = u_data - torch.mean(u_data)
    return u_data


def solve_fast_diag(
    v: GridVariableVector,
    q0: Optional[GridVariable] = None,
    pressure_bc: Optional[grids.ConstantBoundaryConditions] = None,
    implementation: Optional[str] = None,
) -> GridArray:
    """Solve for pressure using the fast diagonalization approach.

      To support backward compatibility, if the pressure_bc are not provided and
      velocity has all periodic boundaries, pressure_bc are assigned to be periodic.

      Args:
          v: a tuple of velocity values for each direction.
          q0: the starting guess for the pressure.
          pressure_bc: the boundary condition to assign to pressure. If None,
          boundary condition is infered from velocity.
          implementation: how to implement fast diagonalization.
          For non-periodic BCs will automatically be matmul.

    Returns:
      A solution to the PPE equation.
    """
    del q0  # unused
    if pressure_bc is None:
        pressure_bc = grids.get_pressure_bc_from_velocity(v)
    if grids.has_all_periodic_boundary_conditions(*v):
        circulant = True
    else:
        circulant = False
        # only matmul implementation supports non-circulant matrices
        implementation = "matmul"
    grid = grids.consistent_grid(*v)
    rhs = fd.divergence(v)
    laplacians = list(map(fd.laplacian_matrix, grid.shape, grid.step))
    laplacians = [lap.to(grid.device) for lap in laplacians]
    rhs_transformed = _rhs_transform(rhs, pressure_bc)
    pinv = solver.pseudoinverse(
        rhs_transformed,
        laplacians,
        rhs_transformed.dtype,
        hermitian=True,
        circulant=circulant,
        implementation=implementation,
    )
    return GridArray(pinv, rhs.offset, rhs.grid)


def projection(
    v: GridVariableVector,
    solve: Callable = solve_fast_diag,
) -> GridVariableVector:
    """
    Apply pressure projection (a discrete Helmholtz decomposition)
    to make a velocity field divergence free.

    Note by S.Cao: this was originally implemented by the jax-cfd team
    but using FDM results having a non-negligible error in fp32.
    One resolution is to use fp64 then cast back to fp32.
    """
    grid = grids.consistent_grid(*v)
    pressure_bc = grids.get_pressure_bc_from_velocity(v)

    q0 = GridArray(torch.zeros(grid.shape), grid.cell_center, grid)
    q0 = pressure_bc.impose_bc(q0)

    q = solve(v, q0, pressure_bc)
    q = pressure_bc.impose_bc(q)
    q_grad = fd.forward_difference(q)
    v_projected = tuple(u.bc.impose_bc(u.array - q_g) for u, q_g in zip(v, q_grad))
    return v_projected



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Modifications copyright (C) 2024 S.Cao
# ported Google's Jax-CFD functional template to PyTorch's tensor ops

"""Prepare initial conditions for simulations."""
import math
from typing import Callable, Optional, Sequence

import torch
import torch.fft as fft

from . import grids, pressure

Array = torch.Tensor
GridArray = grids.GridArray
GridArrayVector = grids.GridArrayVector
GridVariable = grids.GridVariable
GridVariableVector = grids.GridVariableVector
BoundaryConditions = grids.BoundaryConditions


def wrap_velocities(
    v: Sequence[Array],
    grid: grids.Grid,
    bcs: Sequence[BoundaryConditions],
    device: Optional[torch.device] = None,
) -> GridVariableVector:
    """Wrap velocity arrays for input into simulations."""
    device = grid.device if device is None else device
    return tuple(
        GridVariable(GridArray(u, offset, grid).to(device), bc)
        for u, offset, bc in zip(v, grid.cell_faces, bcs)
    )


def wrap_vorticity(
    w: Array,
    grid: grids.Grid,
    bc: BoundaryConditions,
    device: Optional[torch.device] = None,
) -> GridVariable:
    """Wrap vorticity arrays for input into simulations."""
    device = grid.device if device is None else device
    return GridVariable(GridArray(w, grid.cell_faces, grid).to(device), bc)


def _log_normal_density(k, mode: float, variance=0.25):
    """
    Unscaled PDF for a log normal given `mode` and log variance 1.


    """
    mean = math.log(mode) + variance
    logk = torch.log(k)
    return torch.exp(-((mean - logk) ** 2) / 2 / variance - logk)


def McWilliams_density(k, mode: float, tau: float = 1.0):
    """Implements the McWilliams spectral density function.
    |\psi|^2 \sim k^{-1}(tau^2 + (k/k_0)^4)^{-1}
    k_0 is a prescribed wavenumber that the energy peaks.
    tau flattens the spectrum density at low wavenumbers to be bigger.

    Refs:
      McWilliams, J. C. (1984). The emergence of isolated coherent vortices in turbulent flow.
    """
    return (k * (tau**2 + (k / mode) ** 4)) ** (-1)


def _angular_frequency_magnitude(grid: grids.Grid) -> Array:
    frequencies = [
        2 * torch.pi * fft.fftfreq(size, step)
        for size, step in zip(grid.shape, grid.step)
    ]
    freq_vector = torch.stack(torch.meshgrid(*frequencies, indexing="ij"), axis=0)
    return torch.linalg.norm(freq_vector, axis=0)


def spectral_filter(
    spectral_density: Callable[[Array], Array],
    v: Array,
    grid: grids.Grid,
) -> Array:
    """Filter an Array with white noise to match a prescribed spectral density."""
    k = _angular_frequency_magnitude(grid)
    filters = torch.where(k > 0, spectral_density(k), 0.0)
    # The output signal can safely be assumed to be real if our input signal was
    # real, because our spectral density only depends on norm(k).
    return fft.ifftn(fft.fftn(v) * filters).real


def streamfunc_normalize(k, psi):
    nx, ny = psi.shape
    psih = fft.fft2(psi)
    uh_mag = k * psih
    kinetic_energy = (2 * uh_mag.abs() ** 2 / (nx * ny) ** 2).sum()
    return psi / kinetic_energy.sqrt()


def project_and_normalize(
    v: GridVariableVector, maximum_velocity: float = 1
) -> GridVariableVector:
    v = pressure.projection(v)
    vmax = torch.linalg.norm(torch.stack([u.data for u in v]), dim=0).max()
    v = tuple(GridVariable(maximum_velocity * u.array / vmax, u.bc) for u in v)
    return v


def filtered_velocity_field(
    grid: grids.Grid,
    maximum_velocity: float = 1,
    peak_wavenumber: float = 3,
    iterations: int = 3,
    random_state: int = 0,
) -> GridArray:
    """Create divergence-free velocity fields with appropriate spectral filtering.

    Args:
      rng_key: key for seeding the random initial velocity field.
      grid: the grid on which the velocity field is defined.
      maximum_velocity: the maximum speed in the velocity field.
      peak_wavenumber: the velocity field will be filtered so that the largest
        magnitudes are associated with this wavenumber.
      iterations: the number of repeated pressure projection and normalization
        iterations to apply.
    Returns:
      A divergence free velocity field with the given maximum velocity. Associates
      periodic boundary conditions with the velocity field components.
    """

    # Log normal distribution peaked at `peak_wavenumber`. Note that we have to
    # divide by `k ** (ndim - 1)` to account for the volume of the
    # `ndim - 1`-sphere of values with wavenumber `k`.
    def spectral_density(k):
        return _log_normal_density(k, peak_wavenumber) / k ** (grid.ndim - 1)

    random_states = [random_state + i for i in range(grid.ndim)]
    rng = torch.Generator()
    velocity_components = []
    boundary_conditions = []
    for k in random_states:
        rng.manual_seed(k)
        noise = torch.randn(grid.shape, generator=rng)
        velocity_components.append(spectral_filter(spectral_density, noise, grid))
        boundary_conditions.append(grids.periodic_boundary_conditions(grid.ndim))
    velocity = wrap_velocities(velocity_components, grid, boundary_conditions)

    # Due to numerical precision issues, we repeatedly normalize and project the
    # velocity field. This ensures that it is divergence-free and achieves the
    # specified maximum velocity.
    # velocity is ((n, n), (n, n)) GridVariableVector
    for _ in range(iterations):
        velocity = project_and_normalize(velocity, maximum_velocity)
    return velocity


def vorticity_field(
    grid: grids.Grid,
    peak_wavenumber: float = 3,
    random_state: int = 0,
) -> GridArray:
    """Create vorticity field with a spectral filtering
    using the McWilliams power spectrum density function.

    Args:
      rng_key: key for seeding the random initial vorticity field.
      grid: the grid on which the vorticity field is defined.
      peak_wavenumber: the velocity field will be filtered so that the largest
        magnitudes are associated with this wavenumber.

    Returns:
      A vorticity field with periodic boundary condition.
    """

    def spectral_density(k):
        return McWilliams_density(k, peak_wavenumber)

    rng = torch.Generator()
    rng.manual_seed(random_state)
    noise = torch.randn(grid.shape, generator=rng)
    k = _angular_frequency_magnitude(grid)
    psi = spectral_filter(spectral_density, noise, grid)
    psi = streamfunc_normalize(k, psi)
    vorticity = fft.ifftn(fft.fftn(psi) * k**2).real
    boundary_condition = grids.periodic_boundary_conditions(grid.ndim)
    vorticity = wrap_vorticity(vorticity, grid, boundary_condition)

    return vorticity



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Modifications copyright (C) 2024 S.Cao
# ported Google's Jax-CFD functional template to PyTorch's tensor ops

from typing import Optional, Tuple, Union

import torch
import torch.nn as nn

from . import grids

Array = torch.Tensor
Grid = grids.Grid
GridArray = grids.GridArray


def forcing_eval(eval_func):
    def wrapper(
        cls, grid: Grid, field: Optional[Union[Tuple[Array, Array], Array]]
    ) -> Array:
        return eval_func(grid, field)

    return wrapper


class ForcingFn(nn.Module):
    """
    A meta class for forcing functions

    Args:
    vorticity: whether the forcing function is a vorticity forcing
    """

    def __init__(
        self,
        grid: Grid,
        scale: float = 1,
        k: int = 1,
        diam: float = 1.0,
        swap_xy: bool = False,
        vorticity: bool = False,
        offsets: Optional[Tuple[Tuple[float, ...], ...]] = None,
        device: Optional[torch.device] = None,
        **kwargs,
    ):
        super().__init__()
        self.grid = grid
        self.scale = scale
        self.k = k
        self.diam = diam
        self.swap_xy = swap_xy
        self.vorticity = vorticity
        self.offsets = grid.cell_faces if offsets is None else offsets
        self.device = grid.device if device is None else device

    @forcing_eval
    def velocity_eval(grid: Grid, velocity: Optional[Tuple[Array, Array]]) -> Array:
        raise NotImplementedError

    @forcing_eval
    def vorticity_eval(grid: Grid, vorticity: Optional[Array]) -> Array:
        raise NotImplementedError

    def forward(
        self,
        grid: Optional[Grid],
        velocity: Optional[Tuple[Array, Array]] = None,
        vorticity: Optional[Array] = None,
    ) -> Tuple[Array, Array]:
        if not self.vorticity:
            return self.velocity_eval(grid, velocity)
        else:
            return self.vorticity_eval(grid, vorticity)


class KolmogorovForcing(ForcingFn):
    """
    The Kolmogorov forcing function used in
    Sets up the flow that is used in Kochkov et al. [1].
    which is based on Boffetta et al. [2].

    Note in the port: this forcing belongs a larger class
    of isotropic turbulence. See [3].

    References:
    [1] Machine learning-accelerated computational fluid dynamics. Dmitrii
    Kochkov, Jamie A. Smith, Ayya Alieva, Qing Wang, Michael P. Brenner, Stephan
    Hoyer Proceedings of the National Academy of Sciences May 2021, 118 (21)
    e2101784118; DOI: 10.1073/pnas.2101784118.
    https://doi.org/10.1073/pnas.2101784118

    [2] Boffetta, Guido, and Robert E. Ecke. "Two-dimensional turbulence."
    Annual review of fluid mechanics 44 (2012): 427-451.
    https://doi.org/10.1146/annurev-fluid-120710-101240

    [3] McWilliams, J. C. (1984). "The emergence of isolated coherent vortices
    in turbulent flow". Journal of Fluid Mechanics, 146, 21-43.
    """

    def __init__(
        self,
        diam=2 * torch.pi,
        offsets=((0, 0), (0, 0)),
        vorticity=False,
        *args,
        **kwargs,
    ):
        super().__init__(
            *args,
            diam=diam,
            offsets=offsets,
            vorticity=vorticity,
            **kwargs,
        )

    def velocity_eval(
        self,
        grid: Optional[Grid],
        velocity: Optional[Tuple[Array, Array]] = None,
    ) -> Tuple[Array, Array]:
        offsets = self.offsets
        grid = self.grid if grid is None else grid
        domain_factor = 2 * torch.pi / self.diam

        if self.swap_xy:
            x = grid.mesh(offsets[1])[0]
            v = GridArray(
                self.scale * torch.sin(self.k * domain_factor * x), offsets[1], grid
            )
            u = GridArray(torch.zeros_like(v.data), (1, 1 / 2), grid)
            f = (u, v)
        else:
            y = grid.mesh(offsets[0])[1]
            u = GridArray(
                self.scale * torch.sin(self.k * domain_factor * y), offsets[0], grid
            )
            v = GridArray(torch.zeros_like(u.data), (1 / 2, 1), grid)
            f = (u, v)
        return f

    def vorticity_eval(
        self,
        grid: Optional[Grid],
        vorticity: Optional[Array] = None,
    ) -> Array:
        offsets = self.offsets
        grid = self.grid if grid is None else grid
        domain_factor = 2 * torch.pi / self.diam

        if self.swap_xy:
            x = grid.mesh(offsets[1])[0]
            w = GridArray(
                -self.scale
                * self.k
                * domain_factor
                * torch.cos(self.k * domain_factor * x),
                offsets[1],
                grid,
            )
        else:
            y = grid.mesh(offsets[0])[1]
            w = GridArray(
                -self.scale
                * self.k
                * domain_factor
                * torch.cos(self.k * domain_factor * y),
                offsets[0],
                grid,
            )
        return w


def scalar_potential(potential_func):
    def wrapper(cls, x: Array, y: Array, s: float, k: float) -> Array:
        return potential_func(x, y, s, k)

    return wrapper


class SimpleSolenoidalForcing(ForcingFn):
    """
    A simple solenoidal (rotating, divergence free) forcing function template.
    The template forcing is F = (-psi, psi) such that

    Args:
    grid: grid on which to simulate the flow
    scale: a in the equation above, amplitude of the forcing
    k: k in the equation above, wavenumber of the forcing
    """

    def __init__(
        self,
        scale=1,
        diam=1.0,
        k=1.0,
        offsets=((0, 0), (0, 0)),
        vorticity=True,
        *args,
        **kwargs,
    ):
        super().__init__(
            *args,
            scale=scale,
            diam=diam,
            k=k,
            offsets=offsets,
            vorticity=vorticity,
            **kwargs,
        )

    @scalar_potential
    def potential(*args, **kwargs) -> Array:
        raise NotImplementedError

    @scalar_potential
    def vort_potential(*args, **kwargs) -> Array:
        raise NotImplementedError

    def velocity_eval(
        self,
        grid: Optional[Grid],
        velocity: Optional[Tuple[Array, Array]] = None,
    ) -> Tuple[Array, Array]:
        offsets = self.offsets
        grid = self.grid if grid is None else grid
        domain_factor = 2 * torch.pi / self.diam
        k = self.k * domain_factor
        scale = 0.5 * self.scale / (2 * torch.pi) / self.k

        if self.swap_xy:
            x = grid.mesh(offsets[1])[0]
            y = grid.mesh(offsets[0])[1]
            rot = self.potential(x, y, scale, k)
            v = GridArray(rot, offsets[1], grid)
            u = GridArray(-rot, (1, 1 / 2), grid)
            f = (u, v)
        else:
            x = grid.mesh(offsets[0])[0]
            y = grid.mesh(offsets[1])[1]
            rot = self.potential(x, y, scale, k)
            u = GridArray(rot, offsets[0], grid)
            v = GridArray(-rot, (1 / 2, 1), grid)
            f = (u, v)
        return f

    def vorticity_eval(
        self,
        grid: Optional[Grid],
        vorticity: Optional[Array] = None,
    ) -> Array:
        offsets = self.offsets
        grid = self.grid if grid is None else grid
        domain_factor = 2 * torch.pi / self.diam
        k = self.k * domain_factor
        scale = self.scale

        if self.swap_xy:
            x = grid.mesh(offsets[1])[0]
            y = grid.mesh(offsets[0])[1]
            return self.vort_potential(x, y, scale, k)
        else:
            x = grid.mesh(offsets[0])[0]
            y = grid.mesh(offsets[1])[1]
            return self.vort_potential(x, y, scale, k)


class SinCosForcing(SimpleSolenoidalForcing):
    """
    The solenoidal (divergence free) forcing function used in [4].

    Note: in the vorticity-streamfunction formulation, the forcing
    is actually the curl of the velocity field, which
    is a*(sin(2*pi*k*(x+y)) + cos(2*pi*k*(x+y)))
    a=0.1, k=1 in [4]

    References:
    [4] Li, Zongyi, et al. "Fourier Neural Operator for
    Parametric Partial Differential Equations."
    ICLR. 2020.

    Args:
    grid: grid on which to simulate the flow
    scale: a in the equation above, amplitude of the forcing
    k: k in the equation above, wavenumber of the forcing
    """

    def __init__(
        self,
        scale=0.1,
        diam=1.0,
        k=1.0,
        offsets=((0, 0), (0, 0)),
        *args,
        **kwargs,
    ):
        super().__init__(
            *args,
            scale=scale,
            diam=diam,
            k=k,
            offsets=offsets,
            **kwargs,
        )

    @scalar_potential
    def potential(x: Array, y: Array, s: float, k: float) -> Array:
        return s * (torch.sin(k * (x + y)) - torch.cos(k * (x + y)))

    @scalar_potential
    def vort_potential(x: Array, y: Array, s: float, k: float) -> Array:
        return s * (torch.cos(k * (x + y)) + torch.sin(k * (x + y)))



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Modifications copyright (C) 2024 S.Cao
# ported Google's Jax-CFD functional template to PyTorch's tensor ops

from __future__ import annotations

from typing import Any, Callable, List, Sequence, Tuple, Union

import torch
import torch.utils._pytree as pytree

Array = torch.Tensor

def _normalize_axis(axis: int, ndim: int) -> int:
    """Validates and returns positive `axis` value."""
    if not -ndim <= axis < ndim:
        raise ValueError(f"invalid axis {axis} for ndim {ndim}")
    if axis < 0:
        axis += ndim
    return axis


def slice_along_axis(
    inputs: Any, axis: int, idx: Union[slice, int], expect_same_dims: bool = True
) -> Any:
    """Returns slice of `inputs` defined by `idx` along axis `axis`.

    Args:
      inputs: array or a tuple of arrays to slice.
      axis: axis along which to slice the `inputs`.
      idx: index or slice along axis `axis` that is returned.
      expect_same_dims: whether all arrays should have same number of dimensions.

    Returns:
      Slice of `inputs` defined by `idx` along axis `axis`.
    """
    arrays, tree_def = pytree.tree_flatten(inputs)
    ndims = set(a.ndim for a in arrays)
    if expect_same_dims and len(ndims) != 1:
        raise ValueError(
            "arrays in `inputs` expected to have same ndims, but have "
            f"{ndims}. To allow this, pass expect_same_dims=False"
        )
    sliced = []
    for array in arrays:
        ndim = array.ndim
        slc = tuple(
            idx if j == _normalize_axis(axis, ndim) else slice(None)
            for j in range(ndim)
        )
        sliced.append(array[slc])
    return pytree.tree_unflatten(tree_def, sliced)


def split_along_axis(
    inputs: Any, split_idx: int, axis: int, expect_same_dims: bool = True
) -> Tuple[Any, Any]:
    """Returns a tuple of slices of `inputs` split along `axis` at `split_idx`.

    Args:
      inputs: pytree of arrays to split.
      split_idx: index along `axis` where the second split starts.
      axis: axis along which to split the `inputs`.
      expect_same_dims: whether all arrays should have same number of dimensions.

    Returns:
      Tuple of slices of `inputs` split along `axis` at `split_idx`.
    """

    first_slice = slice_along_axis(inputs, axis, slice(0, split_idx), expect_same_dims)
    second_slice = slice_along_axis(
        inputs, axis, slice(split_idx, None), expect_same_dims
    )
    return first_slice, second_slice


def split_axis(inputs: Any, dim: int, keep_dims: bool = False) -> Tuple[Any, ...]:
    """Splits the arrays in `inputs` along `axis`.

    Args:
      inputs: pytree to be split.
      axis: axis along which to split the `inputs`.
      keep_dims: whether to keep `axis` dimension.

    Returns:
      Tuple of pytrees that correspond to slices of `inputs` along `axis`. The
      `axis` dimension is removed if `squeeze is set to True.

    Raises:
      ValueError: if arrays in `inputs` don't have unique size along `axis`.
    """
    arrays, tree_def = pytree.tree_flatten(inputs)
    axis_shapes = set(a.shape[dim] for a in arrays)
    if len(axis_shapes) != 1:
        raise ValueError(f"Arrays must have equal sized axis but got {axis_shapes}")
    (axis_shape,) = axis_shapes
    splits = [torch.split(a, axis_shape, dim=dim) for a in arrays]
    if not keep_dims:
        splits = pytree.tree_map(lambda a: torch.squeeze(a, dim), splits)
    splits = zip(*splits)
    return tuple(pytree.tree_unflatten(tree_def, leaves) for leaves in splits)





# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Modifications copyright (C) 2024 S.Cao
# ported Google's Jax-CFD functional template to PyTorch's tensor ops

from __future__ import annotations

import dataclasses
import math
import numbers
import operator
from functools import partial, reduce

from typing import Any, Callable, Optional, Sequence, Tuple, Union

import numpy as np
import torch
import torch.fft as fft
import torch.nn as nn
import torch.nn.functional as F
import torch.utils._pytree as pytree

from .tensor_utils import split_along_axis

Array = torch.Tensor


class BCType:
    PERIODIC = "periodic"
    DIRICHLET = "dirichlet"
    NEUMANN = "neumann"


@dataclasses.dataclass(init=False, frozen=True)
class Grid:
    """
    port from jax_cfd.base.grids.Grid to pytorch
    Describes the size and shape for an Arakawa C-Grid.

    See https://en.wikipedia.org/wiki/Arakawa_grids.

    This class describes domains that can be written as an outer-product of 1D
    grids. Along each dimension `i`:
    - `shape[i]` gives the whole number of grid cells on a single device.
    - `step[i]` is the width of each grid cell.
    - `(lower, upper) = domain[i]` gives the locations of lower and upper
      boundaries. The identity `upper - lower = step[i] * shape[i]` is enforced.
    """

    shape: Tuple[int, ...]
    step: Tuple[float, ...]
    domain: Tuple[Tuple[float, float], ...]

    def __init__(
        self,
        shape: Sequence[int],
        step: Optional[Union[float, Sequence[float]]] = None,
        domain: Optional[Union[float, Sequence[Tuple[float, float]]]] = None,
        device: Optional[torch.device] = "cpu",
    ):
        super().__init__()
        """Construct a grid object."""
        shape = tuple(operator.index(s) for s in shape)
        object.__setattr__(self, "shape", shape)

        if step is not None and domain is not None:
            raise TypeError("cannot provide both step and domain")
        elif domain is not None:
            if isinstance(domain, (int, float)):
                domain = ((0, domain),) * len(shape)
            else:
                if len(domain) != self.ndim:
                    raise ValueError(
                        "length of domain does not match ndim: "
                        f"{len(domain)} != {self.ndim}"
                    )
                for bounds in domain:
                    if len(bounds) != 2:
                        raise ValueError(
                            f"domain is not sequence of pairs of numbers: {domain}"
                        )
            domain = tuple((float(lower), float(upper)) for lower, upper in domain)
        else:
            if step is None:
                step = 1
            if isinstance(step, numbers.Number):
                step = (step,) * self.ndim
            elif len(step) != self.ndim:
                raise ValueError(
                    "length of step does not match ndim: " f"{len(step)} != {self.ndim}"
                )
            domain = tuple(
                (0.0, float(step_ * size)) for step_, size in zip(step, shape)
            )

        object.__setattr__(self, "domain", domain)

        step = tuple(
            (upper - lower) / size for (lower, upper), size in zip(domain, shape)
        )
        object.__setattr__(self, "step", step)
        object.__setattr__(self, "device", device)

    @property
    def ndim(self) -> int:
        """Returns the number of dimensions of this grid."""
        return len(self.shape)

    @property
    def cell_center(self) -> Tuple[float, ...]:
        """Offset at the center of each grid cell."""
        return self.ndim * (0.5,)

    @property
    def cell_faces(self) -> Tuple[Tuple[float, ...]]:
        """Returns the offsets at each of the 'forward' cell faces."""
        d = self.ndim
        offsets = (torch.eye(d) + torch.ones([d, d])) / 2.0
        return tuple(tuple(float(o) for o in offset) for offset in offsets)

    def stagger(self, v: Tuple[Array, ...]) -> Tuple[Array, ...]:
        """Places the velocity components of `v` on the `Grid`'s cell faces."""
        offsets = self.cell_faces
        return tuple(GridArray(u, o, self) for u, o in zip(v, offsets))

    def center(self, v: Tuple[Array, ...]) -> Tuple[Array, ...]:
        """Places all arrays in the pytree `v` at the `Grid`'s cell center."""
        offset = self.cell_center
        return lambda u: GridArray(u, offset, self), v

    def axes(self, offset: Optional[Sequence[float]] = None) -> Tuple[Array, ...]:
        """Returns a tuple of arrays containing the grid points along each axis.

        Args:
          offset: an optional sequence of length `ndim`. The grid will be shifted by
            `offset * self.step`.

        Returns:
          An tuple of `self.ndim` arrays. The jth return value has shape
          `[self.shape[j]]`.
        """
        if offset is None:
            offset = self.cell_center
        if len(offset) != self.ndim:
            raise ValueError(
                f"unexpected offset length: {len(offset)} vs " f"{self.ndim}"
            )
        return tuple(
            lower + (torch.arange(length) + offset_i) * step
            for (lower, _), offset_i, length, step in zip(
                self.domain, offset, self.shape, self.step
            )
        )

    def fft_axes(self) -> Tuple[Array, ...]:
        """Returns the ordinal frequencies corresponding to the axes.

        Transforms each axis into the *ordinal* frequencies for the Fast Fourier
        Transform (FFT). Multiply by `2 * jnp.pi` to get angular frequencies.

        Returns:
          A tuple of `self.ndim` arrays. The jth return value has shape
          `[self.shape[j]]`.
        """
        freq_axes = tuple(fft.fftfreq(n, d=s) for (n, s) in zip(self.shape, self.step))
        return freq_axes

    def mesh(
        self,
        offset: Optional[Sequence[float]] = None,
    ) -> Tuple[Array, ...]:
        """Returns an tuple of arrays containing positions in each grid cell.

        Args:
          offset: an optional sequence of length `ndim`. The grid will be shifted by
            `offset * self.step`.

        Returns:
          An tuple of `self.ndim` arrays, each of shape `self.shape`. In 3
          dimensions, entry `self.mesh[n][i, j, k]` is the location of point
          `i, j, k` in dimension `n`.
        """
        axes = self.axes(offset)
        x, y = torch.meshgrid(*axes, indexing="ij")
        return x.to(self.device), y.to(self.device)

    def fft_mesh(self) -> Tuple[Array, ...]:
        """Returns a tuple of arrays containing positions in Fourier space."""
        fft_axes = self.fft_axes()
        kx, ky = torch.meshgrid(*fft_axes, indexing="ij")
        return kx.to(self.device), ky.to(self.device)

    def rfft_mesh(self) -> Tuple[Array, ...]:
        """Returns a tuple of arrays containing positions in rfft space."""
        fft_mesh = self.fft_mesh()
        k_max = math.floor(self.shape[-1] / 2.0)
        return tuple(fmesh[..., : k_max + 1] for fmesh in fft_mesh)

    def eval_on_mesh(
        self, fn: Callable[..., Array], offset: Optional[Sequence[float]] = None
    ) -> Array:
        """Evaluates the function on the grid mesh with the specified offset.

        Args:
          fn: A function that accepts the mesh arrays and returns an array.
          offset: an optional sequence of length `ndim`.  If not specified, uses the
            offset for the cell center.

        Returns:
          fn(x, y, ...) evaluated on the mesh, as a GridArray with specified offset.
        """
        if offset is None:
            offset = self.cell_center
        return GridArray(fn(*self.mesh(offset)), offset, self)


@dataclasses.dataclass
class GridArray(torch.Tensor):
    """
    the original jax implentation uses np.lib.mixins.NDArrayOperatorsMixin
    and the __array_ufunc__ method to implement arithmetic operations
    here it is modified to use torch.Tensor as the base class
    and __torch_function__ to do various things like clone() and to()
    reference: https://pytorch.org/docs/stable/notes/extending.html

    Data with an alignment offset and an associated grid.

    Offset values in the range [0, 1] fall within a single grid cell.

    Examples:
      offset=(0, 0) means that each point is at the bottom-left corner.
      offset=(0.5, 0.5) is at the grid center.
      offset=(1, 0.5) is centered on the right-side edge.

    Attributes:
      data: array values.
      offset: alignment location of the data with respect to the grid.
      grid: the Grid associated with the array data.
      dtype: type of the array data.
      shape: lengths of the array dimensions.
    """

    # Don't (yet) enforce any explicit consistency requirements between data.ndim
    # and len(offset), e.g., so we can feel to add extra time/batch/channel
    # dimensions. But in most cases they should probably match.
    # Also don't enforce explicit consistency between data.shape and grid.shape,
    # but similarly they should probably match.

    data: Array = None
    offset: Tuple[float, ...] = None
    grid: Grid = None

    def tree_flatten(self):
        """Returns flattening recipe for GridArray JAX pytree."""
        children = (self.data,)
        aux_data = (self.offset, self.grid)
        return children, aux_data

    @staticmethod
    def __new__(cls, x, offset, grid, *args, **kwargs):
        return super().__new__(cls, x, *args, **kwargs)

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        """Returns unflattening recipe for GridArray JAX pytree."""
        return cls(*children, *aux_data)

    @property
    def dtype(self):
        return self.data.dtype

    @property
    def shape(self) -> Tuple[int, ...]:
        return self.data.shape

    def clone(self, *args, **kwargs):
        return super().clone(*args, **kwargs)

    def to(self, *args, **kwargs):
        return GridArray(self.data.to(*args, **kwargs), self.offset, self.grid)

    _HANDLED_TYPES = (numbers.Number, Array)

    @classmethod
    def __torch_function__(self, ufunc, types, args=(), kwargs=None):
        """Define arithmetic on GridArrays using NumPy's mixin."""
        if kwargs is None:
            kwargs = {}
        if not all(issubclass(t, self._HANDLED_TYPES + (GridArray,)) for t in types):
            return NotImplemented
        try:
            # get the corresponding torch function to the NumPy ufunc
            func = getattr(torch, ufunc.__name__)
        except AttributeError:
            return NotImplemented
        arrays = [x.data if isinstance(x, GridArray) else x for x in args]
        result = func(*arrays)
        offset = consistent_offset(*[x for x in args if isinstance(x, GridArray)])
        grid = consistent_grid(*[x for x in args if isinstance(x, GridArray)])
        if isinstance(result, tuple):
            return tuple(GridArray(r, offset, grid) for r in result)
        else:
            return GridArray(result, offset, grid)


GridArrayVector = Tuple[GridArray, ...]


@dataclasses.dataclass
class GridVariable:
    """Associates a GridArray with BoundaryConditions.

    Performing pad and shift operations, e.g. for finite difference calculations,
    requires boundary condition (BC) information. Since different variables in a
    PDE system can have different BCs, this class associates a specific variable's
    data with its BCs.

    Array operations on GridVariables act like array operations on the
    encapsulated GridArray.

    Attributes:
      array: GridArray with the array data, offset, and associated grid.
      bc: boundary conditions for this variable.
      grid: the Grid associated with the array data.
      dtype: type of the array data.
      shape: lengths of the array dimensions.
      data: array values.
      offset: alignment location of the data with respect to the grid.
      grid: the Grid associated with the array data.
    """

    array: GridArray
    bc: BoundaryConditions

    def __post_init__(self):
        if not isinstance(self.array, GridArray):  # frequently missed by pytype
            raise ValueError(
                f"Expected array type to be GridArray, got {type(self.array)}"
            )
        if len(self.bc.types) != self.grid.ndim:
            raise ValueError(
                "Incompatible dimension between grid and bc, grid dimension = "
                f"{self.grid.ndim}, bc dimension = {len(self.bc.types)}"
            )

    def tree_flatten(self):
        """Returns flattening recipe for GridVariable JAX pytree."""
        children = (self.array,)
        aux_data = (self.bc,)
        return children, aux_data

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        """Returns unflattening recipe for GridVariable JAX pytree."""
        return cls(*children, *aux_data)

    @property
    def dtype(self):
        return self.array.dtype

    @property
    def shape(self) -> Tuple[int, ...]:
        return self.array.shape

    @property
    def data(self) -> Array:
        return self.array.data

    @property
    def offset(self) -> Tuple[float, ...]:
        return self.array.offset

    @property
    def grid(self) -> Grid:
        return self.array.grid

    def shift(
        self,
        offset: int,
        dim: int,
    ) -> GridArray:
        """Shift this GridVariable by `offset`.

        Args:
          offset: positive or negative integer offset to shift.
          dim: axis to shift along.

        Returns:
          A copy of the encapsulated GridArray, shifted by `offset`. The returned
          GridArray has offset `u.offset + offset`.
        """
        return self.bc.shift(self.array, offset, dim)

    def _interior_grid(self) -> Grid:
        """Returns only the interior grid points."""
        grid = self.array.grid
        domain = list(grid.domain)
        shape = list(grid.shape)
        for axis in range(self.grid.ndim):
            # nothing happens in periodic case
            if self.bc.types[axis][1] == "periodic":
                continue
            # nothing happens if the offset is not 0.0 or 1.0
            # this will automatically set the grid to interior.
            if torch.isclose(self.array.offset[axis], 1.0):
                shape[axis] -= 1
                domain[axis] = (domain[axis][0], domain[axis][1] - grid.step[axis])
            elif torch.isclose(self.array.offset[axis], 0.0):
                shape[axis] -= 1
                domain[axis] = (domain[axis][0] + grid.step[axis], domain[axis][1])
        return Grid(shape, domain=tuple(domain))

    def _interior_array(self) -> Array:
        """Returns only the interior points of self.array."""
        data = self.array.data
        for axis in range(self.grid.ndim):
            # nothing happens in periodic case
            if self.bc.types[axis][1] == "periodic":
                continue
            # nothing happens if the offset is not 0.0 or 1.0
            if torch.isclose(self.offset[axis], 1.0):
                data, _ = split_along_axis(data, -1, axis)
            elif torch.isclose(self.offset[axis], 0.0):
                _, data = split_along_axis(data, 1, axis)

        return data

    def interior(self) -> GridArray:
        """Returns a GridArray associated only with interior points.

         Interior is defined as the following:register
           for d in range(u.grid.ndim):
            points = u.grid.axes(offset=u.offset[d])
            interior_points =
              all points where grid.domain[d][0] < points < grid.domain[d][1]

        The exception is when the boundary conditions are periodic,
        in which case all points are included in the interior.

        In case of dirichlet with edge offset, the grid and array size is reduced,
        since one scalar lies exactly on the boundary. In all other cases,
        self.grid and self.array are returned.
        """
        interior_array = self._interior_array()
        interior_grid = self._interior_grid()
        return GridArray(interior_array, self.array.offset, interior_grid)

    def enforce_edge_bc(self, *args) -> GridVariable:
        """Returns the GridVariable with edge BC enforced, if applicable.

        For GridVariables having nonperiodic BC and offset 0 or 1, there are values
        in the array data that are dependent on the boundary condition.
        enforce_edge_bc() changes these boundary values to match the prescribed BC.

        Args:
          *args: any optional values passed into BoundaryConditions values method.
        """
        if self.grid.shape != self.array.data.shape:
            raise ValueError("Stored array and grid have mismatched sizes.")
        data = torch.tensor(self.array.data)
        for axis in range(self.grid.ndim):
            if "periodic" not in self.bc.types[axis]:
                values = self.bc.values(axis, self.grid, *args)
                for boundary_side in range(2):
                    if torch.isclose(self.array.offset[axis], boundary_side):
                        # boundary data is set to match self.bc:
                        all_slice = [
                            slice(None, None, None),
                        ] * self.grid.ndim
                        all_slice[axis] = -boundary_side
                        data = data.at[tuple(all_slice)].set(values[boundary_side])
        return GridVariable(
            array=GridArray(data, self.array.offset, self.grid), bc=self.bc
        )


GridVariableVector = Tuple[GridVariable, ...]


class GridArrayTensor(Array):
    """A numpy array of GridArrays, representing a physical tensor field.

    Packing tensor coordinates into a numpy array of dtype object is useful
    because pointwise matrix operations like trace, transpose, and matrix
    multiplications of physical tensor quantities is meaningful.

    Example usage:
      grad = fd.gradient_tensor(uv)                    # a rank 2 Tensor
      strain_rate = (grad + grad.T) / 2.
      nu_smag = np.sqrt(np.trace(strain_rate.dot(strain_rate)))
      nu_smag = Tensor(nu_smag)                        # a rank 0 Tensor
      subgrid_stress = -2 * nu_smag * strain_rate      # a rank 2 Tensor
    """

    def __new__(cls, arrays):
        return torch.asarray(arrays).view(cls)


pytree.register_pytree_node(
    GridArrayTensor,
    lambda tensor: (tensor.ravel().tolist(), tensor.shape),
    lambda shape, arrays: GridArrayTensor(torch.asarray(arrays).reshape(shape)),
)


def applied(func):
    """Convert an array function into one defined on GridArrays.

    Since `func` can only act on `data` attribute of GridArray, it implicitly
    enforces that `func` cannot modify the other attributes such as offset.

    Args:
      func: function being wrapped.

    Returns:
      A wrapped version of `func` that takes GridArray instead of Array args.
    """

    def wrapper(*args, **kwargs):  # pylint: disable=missing-docstring
        for arg in args + tuple(kwargs.values()):
            if isinstance(arg, GridVariable):
                raise ValueError("grids.applied() cannot be used with GridVariable")

        offset = consistent_offset(
            *[
                arg
                for arg in args + tuple(kwargs.values())
                if isinstance(arg, GridArray)
            ]
        )
        grid = consistent_grid(
            *[
                arg
                for arg in args + tuple(kwargs.values())
                if isinstance(arg, GridArray)
            ]
        )
        raw_args = [arg.data if isinstance(arg, GridArray) else arg for arg in args]
        raw_kwargs = {
            k: v.data if isinstance(v, GridArray) else v for k, v in kwargs.items()
        }
        data = func(*raw_args, **raw_kwargs)
        return GridArray(data, offset, grid)

    return wrapper


def averaged_offset(*arrays: Union[GridArray, GridVariable]) -> Tuple[float, ...]:
    """Returns the averaged offset of the given arrays."""
    offsets = torch.as_tensor([array.offset for array in arrays])
    offset = torch.mean(offsets, dim=0)
    return tuple(offset.tolist())


def consistent_offset(*arrays: Array) -> Tuple[float, ...]:
    """Returns the unique offset, or raises InconsistentOffsetError."""
    offsets = {array.offset for array in arrays}
    if len(offsets) != 1:
        raise Exception(f"arrays do not have a unique offset: {offsets}")
    (offset,) = offsets
    return offset


def consistent_grid(*arrays: GridArray):
    """Returns the unique grid, or raises InconsistentGridError."""
    grids = {array.grid for array in arrays}
    if len(grids) != 1:
        raise Exception(f"arrays do not have a unique grid: {grids}")
    (grid,) = grids
    return grid


@dataclasses.dataclass(init=False, frozen=True)
class BoundaryConditions:
    """Base class for boundary conditions on a PDE variable.

    Attributes:
      types: `types[i]` is a tuple specifying the lower and upper BC types for
        dimension `i`.
    """

    types: Tuple[Tuple[str, str], ...]

    def shift(
        self,
        u: GridArray,
        offset: int,
        dim: int,
    ) -> GridArray:
        """Shift an GridArray by `offset`.

        Args:
          u: an `GridArray` object.
          offset: positive or negative integer offset to shift.
          dim: axis to shift along.

        Returns:
          A copy of `u`, shifted by `offset`. The returned `GridArray` has offset
          `u.offset + offset`.
        """
        raise NotImplementedError(
            "shift() not implemented in BoundaryConditions base class."
        )

    def values(
        self,
        dim: int,
        grid: Grid,
        offset: Optional[Tuple[float, ...]],
        time: Optional[float],
    ) -> Tuple[Optional[Array], Optional[Array]]:
        """Returns Arrays specifying boundary values on the grid along axis.

        Args:
          dim: axis along which to return boundary values.
          grid: a `Grid` object on which to evaluate boundary conditions.
          offset: a Tuple of offsets that specifies (along with grid) where to
            evaluate boundary conditions in space.
          time: a float used as an input to boundary function.

        Returns:
          A tuple of arrays of grid.ndim - 1 dimensions that specify values on the
          boundary. In case of periodic boundaries, returns a tuple(None,None).
        """
        raise NotImplementedError(
            "values() not implemented in BoundaryConditions base class."
        )


@dataclasses.dataclass(init=False, frozen=True)
class ConstantBoundaryConditions(BoundaryConditions):
    """Boundary conditions for a PDE variable that are constant in space and time.

    Example usage:
      grid = Grid((10, 10))
      array = GridArray(np.zeros((10, 10)), offset=(0.5, 0.5), grid)
      bc = ConstantBoundaryConditions(((BCType.PERIODIC, BCType.PERIODIC),
                                          (BCType.DIRICHLET, BCType.DIRICHLET)),
                                          ((0.0, 10.0),(1.0, 0.0)))
      u = GridVariable(array, bc)

    Attributes:
      types: `types[i]` is a tuple specifying the lower and upper BC types for
        dimension `i`.
    """

    types: Tuple[Tuple[str, str], ...]
    _values: Tuple[Tuple[Optional[float], Optional[float]], ...]

    def __init__(
        self,
        types: Sequence[Tuple[str, str]],
        values: Sequence[Tuple[Optional[float], Optional[float]]],
    ):
        types = tuple(types)
        values = tuple(values)
        object.__setattr__(self, "types", types)
        object.__setattr__(self, "_values", values)

    def shift(
        self,
        u: GridArray,
        offset: int,
        dim: int,
    ) -> GridArray:
        """Shift an GridArray by `offset`.

        Args:
          u: an `GridArray` object.
          offset: positive or negative integer offset to shift.
          dim: axis to shift along.

        Returns:
          A copy of `u`, shifted by `offset`. The returned `GridArray` has offset
          `u.offset + offset`.
        """
        padded = self._pad(u, offset, dim)
        trimmed = self._trim(padded, -offset, dim)
        # print(u.shape, offset)
        # print(padded.shape, trimmed.shape)
        return trimmed

    def _is_aligned(self, u: GridArray, dim: int) -> bool:
        """Checks if array u contains all interior domain information.

        For dirichlet edge aligned boundary, the value that lies exactly on the
        boundary does not have to be specified by u.
        Neumann edge aligned boundary is not defined.

        Args:
        u: array that should contain interior data
        dim: axis along which to check

        Returns:
        True if u is aligned, and raises error otherwise.
        """
        size_diff = u.shape[dim] - u.grid.shape[dim]
        if self.types[dim][0] == BCType.DIRICHLET and np.isclose(u.offset[dim], 1):
            size_diff += 1
        if self.types[dim][1] == BCType.DIRICHLET and np.isclose(u.offset[dim], 1):
            size_diff += 1
        if self.types[dim][0] == BCType.NEUMANN and np.isclose(u.offset[dim] % 1, 0):
            raise NotImplementedError("Edge-aligned neumann BC are not implemented.")
        if size_diff < 0:
            raise ValueError("the GridArray does not contain all interior grid values.")
        return True

    def _pad(
        self,
        u: GridArray,
        width: int,
        dim: int,
    ) -> GridArray:
        """Pad a GridArray by `padding`.

        Important: _pad makes no sense past 1 ghost cell for nonperiodic
        boundaries. This is sufficient for jax_cfd finite difference code.

        Args:
          u: a `GridArray` object.
          width: number of elements to pad along axis. Use negative value for lower
            boundary or positive value for upper boundary.
          dim: axis to pad along.

        Returns:
          Padded array, elongated along the indicated axis.
        """
        if width < 0:  # pad lower boundary
            bc_type = self.types[dim][0]
            padding = (-width, 0)
        else:  # pad upper boundary
            bc_type = self.types[dim][1]
            padding = (0, width)

        full_padding = [(0, 0)] * u.grid.ndim
        full_padding[dim] = padding

        offset = list(u.offset)
        offset[dim] -= padding[0]

        if bc_type != BCType.PERIODIC and abs(width) > 1:
            raise ValueError(
                "Padding past 1 ghost cell is not defined in nonperiodic case."
            )

        if bc_type == BCType.PERIODIC:
            # self.values are ignored here
            pad_kwargs = dict(mode="circular")
        elif bc_type == BCType.DIRICHLET:
            if np.isclose(u.offset[dim] % 1, 0.5):  # cell center
                # make the linearly interpolated value equal to the boundary by setting
                # the padded values to the negative symmetric values
                data = 2 * expand_dims_pad(
                    u.data, full_padding, mode="constant", constant_values=self._values
                ) - expand_dims_pad(u.data, full_padding, mode="reflect")
                return GridArray(data, tuple(offset), u.grid)
            elif np.isclose(u.offset[dim] % 1, 0):  # cell edge
                pad_kwargs = dict(mode="constant", constant_values=self._values)
            else:
                raise ValueError(
                    "expected offset to be an edge or cell center, got "
                    f"offset[axis]={u.offset[dim]}"
                )
        elif bc_type == BCType.NEUMANN:
            if not (
                np.isclose(u.offset[dim] % 1, 0) or np.isclose(u.offset[dim] % 1, 0.5)
            ):
                raise ValueError(
                    "expected offset to be an edge or cell center, got "
                    f"offset[axis]={u.offset[dim]}"
                )
            else:
                # When the data is cell-centered, computes the backward difference.
                # When the data is on cell edges, boundary is set such that
                # (u_last_interior - u_boundary)/grid_step = neumann_bc_value.
                data = expand_dims_pad(
                    u.data, full_padding, mode="replicate"
                ) + u.grid.step[dim] * (
                    expand_dims_pad(u.data, full_padding, mode="constant")
                    - expand_dims_pad(
                        u.data,
                        full_padding,
                        mode="constant",
                        constant_values=self._values,
                    )
                )
                return GridArray(data, tuple(offset), u.grid)

        else:
            raise ValueError("invalid boundary type")
        data = expand_dims_pad(u.data, full_padding, **pad_kwargs)
        return GridArray(data, tuple(offset), u.grid)

    def _trim(
        self,
        u: GridArray,
        width: int,
        dim: int,
    ) -> GridArray:
        """Trim padding from a GridArray.

        Args:
          u: a `GridArray` object.
          width: number of elements to trim along axis. Use negative value for lower
            boundary or positive value for upper boundary.
          dim: axis to trim along.

        Returns:
          Trimmed array, shrunk along the indicated axis.
        """
        if width < 0:  # trim lower boundary
            padding = (-width, 0)
        else:  # trim upper boundary
            padding = (0, width)

        limit_index = u.data.shape[dim] - padding[1]
        data = u.data.index_select(dim=dim, 
                                   index=torch.arange(padding[0], 
                                                      limit_index,
                                                      device=u.data.device))
        offset = list(u.offset)
        offset[dim] += padding[0]
        return GridArray(data, tuple(offset), u.grid)

    def values(self, dim: int, grid: Grid) -> Tuple[Optional[Array], Optional[Array]]:
        """Returns boundary values on the grid along axis.

        Args:
          dim: axis along which to return boundary values.
          grid: a `Grid` object on which to evaluate boundary conditions.

        Returns:
          A tuple of arrays of grid.ndim - 1 dimensions that specify values on the
          boundary. In case of periodic boundaries, returns a tuple(None,None).
        """
        if None in self._values[dim]:
            return (None, None)
        bc = tuple(
            torch.full(grid.shape[:dim] + grid.shape[dim + 1 :], self._values[dim][-i])
            for i in [0, 1]
        )
        return bc

    def _trim_padding(self, u: GridArray, dim: int = 0, trim_side: str = "both"):
        """Trims padding from a GridArray along axis and returns the array interior.

        Args:
        u: a `GridArray` object.
        dim: axis to trim along.
        trim_side: if 'both', trims both sides. If 'right', trims the right side.
            If 'left', the left side.

        Returns:
        Trimmed array, shrunk along the indicated axis side.
        """
        padding = (0, 0)
        if u.shape[dim] >= u.grid.shape[dim]:
            # number of cells that were padded on the left
            negative_trim = 0
            if u.offset[dim] <= 0 and (trim_side == "both" or trim_side == "left"):
                negative_trim = -math.ceil(-u.offset[dim])
                # periodic is a special case. Shifted data might still contain all the
                # information.
                if self.types[dim][0] == BCType.PERIODIC:
                    negative_trim = max(negative_trim, u.grid.shape[dim] - u.shape[dim])
                # for both DIRICHLET and NEUMANN cases the value on grid.domain[0] is
                # a dependent value.
                elif np.isclose(u.offset[dim] % 1, 0):
                    negative_trim -= 1
                u = self._trim(u, negative_trim, dim)
            # number of cells that were padded on the right
            positive_trim = 0
            if trim_side == "right" or trim_side == "both":
                # periodic is a special case. Boundary on one side depends on the other
                # side.
                if self.types[dim][1] == BCType.PERIODIC:
                    positive_trim = max(u.shape[dim] - u.grid.shape[dim], 0)
                else:
                    # for other cases, where to trim depends only on the boundary type
                    # and data offset.
                    last_u_offset = u.shape[dim] + u.offset[dim] - 1
                    boundary_offset = u.grid.shape[dim]
                    if last_u_offset >= boundary_offset:
                        positive_trim = math.ceil(last_u_offset - boundary_offset)
                        if self.types[dim][1] == BCType.DIRICHLET and np.isclose(
                            u.offset[dim] % 1, 0
                        ):
                            positive_trim += 1
        if positive_trim > 0:
            u = self._trim(u, positive_trim, dim)
        # combining existing padding with new padding
        padding = (-negative_trim, positive_trim)
        return u, padding

    def trim_boundary(self, u: GridArray) -> GridArray:
        """Returns GridArray without the grid points on the boundary.

        Some grid points of GridArray might coincide with boundary. This trims those
        values. If the array was padded beforehand, removes the padding.

        Args:
        u: a `GridArray` object.

        Returns:
        A GridArray shrunk along certain dimensions.
        """
        for axis in range(u.grid.ndim):
            _ = self._is_aligned(u, axis)
            u, _ = self._trim_padding(u, axis)
        return u

    def pad_and_impose_bc(
        self,
        u: GridArray,
        offset_to_pad_to: Optional[Tuple[float, ...]] = None,
        mode: Optional[str] = "extend",
    ) -> GridVariable:
        """Returns GridVariable with correct boundary values.

        Some grid points of GridArray might coincide with boundary. This ensures
        that the GridVariable.array agrees with GridVariable.bc.
        Args:
        u: a `GridArray` object that specifies only scalar values on the internal
            nodes.
        offset_to_pad_to: a Tuple of desired offset to pad to. Note that if the
            function is given just an interior array in dirichlet case, it can pad
            to both 0 offset and 1 offset.
        mode: type of padding to use in non-periodic case.
            Mirror mirrors the flow across the boundary.
            Extend extends the last well-defined value past the boundary.

        Returns:
        A GridVariable that has correct boundary values.
        """
        if offset_to_pad_to is None:
            offset_to_pad_to = u.offset
            for axis in range(u.grid.ndim):
                _ = self._is_aligned(u, axis)
                if self.types[axis][0] == BCType.DIRICHLET and np.isclose(
                    u.offset[axis], 1.0
                ):
                    if np.isclose(offset_to_pad_to[axis], 1.0):
                        u = self._pad(u, 1, axis, mode=mode)
                    elif np.isclose(offset_to_pad_to[axis], 0.0):
                        u = self._pad(u, -1, axis, mode=mode)
        return GridVariable(u, self)

    def impose_bc(self, u: GridArray) -> GridVariable:
        """Returns GridVariable with correct boundary condition.

        Some grid points of GridArray might coincide with boundary. This ensures
        that the GridVariable.array agrees with GridVariable.bc.
        Args:
        u: a `GridArray` object.

        Returns:
        A GridVariable that has correct boundary values and is restricted to the
        domain.
        """
        offset = u.offset
        u = self.trim_boundary(u)
        return self.pad_and_impose_bc(u, offset)

    trim = _trim
    pad = _pad


class HomogeneousBoundaryConditions(ConstantBoundaryConditions):
    """Boundary conditions for a PDE variable.

    Example usage:
      grid = Grid((10, 10))
      array = GridArray(np.zeros((10, 10)), offset=(0.5, 0.5), grid)
      bc = ConstantBoundaryConditions(((BCType.PERIODIC, BCType.PERIODIC),
                                          (BCType.DIRICHLET, BCType.DIRICHLET)))
      u = GridVariable(array, bc)

    Attributes:
      types: `types[i]` is a tuple specifying the lower and upper BC types for
        dimension `i`.
    """

    def __init__(self, types: Sequence[Tuple[str, str]]):

        ndim = len(types)
        values = ((0.0, 0.0),) * ndim
        super(HomogeneousBoundaryConditions, self).__init__(types, values)


def is_periodic_boundary_conditions(c: GridVariable, dim: int) -> bool:
    """Returns true if scalar has periodic bc along axis."""
    if c.bc.types[dim][0] != BCType.PERIODIC:
        return False
    return True


# Convenience utilities to ease updating of BoundaryConditions implementation
def periodic_boundary_conditions(ndim: int) -> BoundaryConditions:
    """Returns periodic BCs for a variable with `ndim` spatial dimension."""
    return HomogeneousBoundaryConditions(((BCType.PERIODIC, BCType.PERIODIC),) * ndim)


def consistent_boundary_conditions(*arrays: GridVariable) -> Tuple[str, ...]:
    """Returns whether BCs are periodic.

    Mixed periodic/nonperiodic boundaries along the same boundary do not make
    sense. The function checks that the boundary is either periodic or not and
    throws an error if its mixed.

    Args:
      *arrays: a list of gridvariables.

    Returns:
      a list of types of boundaries corresponding to each axis if
      they are consistent.
    """
    bc_types = []
    for axis in range(arrays[0].grid.ndim):
        bcs = {is_periodic_boundary_conditions(array, axis) for array in arrays}
        if len(bcs) != 1:
            raise Exception(f"arrays do not have consistent bc: {arrays}")
        elif bcs.pop():
            bc_types.append("periodic")
        else:
            bc_types.append("nonperiodic")
    return tuple(bc_types)


def get_pressure_bc_from_velocity(
    v: GridVariableVector,
) -> HomogeneousBoundaryConditions:
    """Returns pressure boundary conditions for the specified velocity."""
    # assumes that if the boundary is not periodic, pressure BC is zero flux.
    velocity_bc_types = consistent_boundary_conditions(*v)
    pressure_bc_types = []
    for velocity_bc_type in velocity_bc_types:
        if velocity_bc_type == "periodic":
            pressure_bc_types.append((BCType.PERIODIC, BCType.PERIODIC))
        else:
            pressure_bc_types.append((BCType.NEUMANN, BCType.NEUMANN))
    return HomogeneousBoundaryConditions(pressure_bc_types)


def has_all_periodic_boundary_conditions(*arrays: GridVariable) -> bool:
    """Returns True if arrays have periodic BC in every dimension, else False."""
    for array in arrays:
        for axis in range(array.grid.ndim):
            if not is_periodic_boundary_conditions(array, axis):
                return False
    return True


def expand_dims_pad(
    inputs: Any,
    pad: Tuple[Tuple[int, int], ...],
    dim: int = 2,
    mode: str = "constant",
    constant_values: float = 0,
) -> Any:
    """
    wrapper for F.pad with a dimension checker
    note: jnp's pad pad_width starts from the first dimension to the last dimension
    while torch's pad pad_width starts from the last dimension to the previous dimension
    example: for torch (1, 1, 2, 2) means padding last dim by (1, 1) and 2nd to last by (2, 2)

    Args:
      inputs: array or a tuple of arrays to pad.
      pad_width: padding width for each dimension.
      mode: padding mode, one of 'constant', 'reflect', 'symmetric'.
      values: constant value to pad with.

    Returns:
      Padded `inputs`.
    """
    assert len(pad) == inputs.ndim, "pad must have the same length as inputs.ndim"
    if not isinstance(inputs, torch.Tensor):
        raise ValueError("inputs must be a torch.Tensor")
    if dim == inputs.ndim:
        inputs = inputs.unsqueeze(0)
    pad = reduce(lambda a, x: x + a, pad, ())  # flatten the pad and reverse the order
    if mode == "constant":
        array = F.pad(inputs, pad, mode=mode, value=constant_values)
    elif mode == "reflect" or mode == "circular":
        array = F.pad(inputs, pad, mode=mode)
    else:
        raise NotImplementedError(f"invalid mode {mode} for torch.nn.functional.pad")

    return array.squeeze(0) if dim != array.ndim else array



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Modifications copyright (C) 2024 S.Cao
# ported Google's Jax-CFD functional template to PyTorch's tensor ops

"""Fast diagonalization method for inverting linear operators."""
from functools import reduce
from typing import Callable, Optional, Sequence, Union, List, Tuple

import numpy as np

import torch
import torch.fft as fft


Array = torch.Tensor


def outer_sum(x: Union[List[Array], Tuple[Array]]) -> Array:
    """
    Returns the outer sum of a list of one dimensional arrays
    Example:
    x = [a, b, c]
    out = a[..., None, None] + b[..., None] + c
    """

    def _sum(a, b):
        return a[..., None] + b

    return reduce(_sum, x)


def transform(
    func: Callable[[Array], Array],
    operators: Sequence[Array],
    dtype: torch.dtype,
    *,
    hermitian: bool = False,
    circulant: bool = False,
    implementation: Optional[str] = None,
) -> Callable[[Array], Array]:
    """Apply a linear operator written as a sum of operators on each axis.

    Such linear operators are *separable*, and can be written as a sum of tensor
    products, e.g., `operators = [A, B]` corresponds to the linear operator
    A ⊗ I + I ⊗ B, where the tensor product ⊗ indicates a separation between
    operators applied along the first and second axis.

    This function computes matrix-valued functions of such linear operators via
    the "fast diagonalization method" [1]:
      F(A ⊗ I + I ⊗ B)
      = (X(A) ⊗ X(B)) F(Λ(A) ⊗ I + I ⊗ Λ(B)) (X(A)^{-1} ⊗ X(B)^{-1})

    where X(A) denotes the matrix of eigenvectors of A and Λ(A) denotes the
    (diagonal) matrix of eigenvalues. The function `F` is easy to compute in
    this basis, because matrix Λ(A) ⊗ I + I ⊗ Λ(B) is diagonal.

    The current implementation directly diagonalizes dense matrices for each
    linear operator, which limits it's applicability to grids with less than
    1e3-1e4 elements per side (~1 second to several minutes of setup time).

    Example: The Laplacian operator can be written as a sum of 1D Laplacian
    operators along each axis, i.e., as a sum of 1D convolutions along each axis.
    This can be seen mathematically (∇² = ∂²/∂x² + ∂²/∂y² + ∂²/∂z²) or by
    decomposing the 2D kernel:

      [0  1  0]               [ 1]
      [1 -4  1] = [1 -2  1] + [-2]
      [0  1  0]               [ 1]

    Args:
      func: NumPy function applied in the diagonal basis that is passed the
        N-dimensional array of eigenvalues (one dimension for each linear
        operator).
      operators: forward linear operators as matrices, applied along each axis.
        Each of these matrices is diagonalized.
      dtype: dtype of the right-hand-side.
      hermitian: whether or not all linear operator are Hermitian (i.e., symmetric
        in the real valued case).
      circulant: whether or not all linear operators are circulant.
      implementation: how to implement fast diagonalization. Default uses 'rfft'
        for grid size larger than 1024 and 'matmul' otherwise:
        - 'matmul': scales like O(N**(d+1)) for d N-dimensional operators, but
          makes good use of matmul hardware. Requires hermitian=True.
        - 'fft': scales like O(N**d * log(N)) for d N-dimensional operators.
          Requires circulant=True.
        - 'rfft': use the RFFT instead of the FFT. This is a little faster than
          'fft' but also has slightly larger error. It currently requires an even
          sized last axis and circulant=True.
      precision: numerical precision for matrix multplication. Only relevant on
        TPUs with implementation='matmul'.

    Returns:
      A function that computes the transformation of the indicated operator.

    References:
    [1] Lynch, R. E., Rice, J. R. & Thomas, D. H. Direct solution of partial
        difference equations by tensor product methods. Numer. Math. 6, 185–199
        (1964). https://paperpile.com/app/p/b7fdea4e-b2f7-0ada-b056-a282325c3ecf
    """
    if any(op.ndim != 2 or op.shape[0] != op.shape[1] for op in operators):
        raise ValueError(
            "operators are not all square matrices. Shapes are "
            + ", ".join(str(op.shape) for op in operators)
        )

    if implementation is None:
        implementation = "rfft"
    if implementation == "rfft" and operators[-1].shape[0] % 2:
        implementation = "matmul"

    if implementation == "matmul":
        if not hermitian:
            raise ValueError(
                "non-hermitian operators not yet supported with "
                'implementation="matmul"'
            )
        return _hermitian_matmul_transform(func, operators, dtype)
    elif implementation == "fft":
        if not circulant:
            raise ValueError(
                "non-circulant operators not yet supported with " 'implementation="fft"'
            )
        return _circulant_fft_transform(func, operators, dtype)
    elif implementation == "rfft":
        if not circulant:
            raise ValueError(
                "non-circulant operators not yet supported with "
                'implementation="rfft"'
            )
        return _circulant_rfft_transform(func, operators, dtype)
    else:
        raise ValueError(f"invalid implementation: {implementation}")


def _hermitian_matmul_transform(
    func: Callable[[Array], Array],
    operators: Sequence[Array],
    dtype: torch.dtype,
) -> Callable[[Array], Array]:
    """Fast diagonalization by matrix multiplication along each axis."""
    eigenvalues, eigenvectors = zip(*map(torch.linalg.eigh, operators))

    # Example: if eigenvalues=[a, b, c], then:
    #   summed_eigenvalues[i, j, k] == a[i] + b[j] + c[k]
    # for all i, j, k.
    summed_eigenvalues = outer_sum(eigenvalues)
    diagonals = torch.asarray(func(summed_eigenvalues), dtype)
    eigenvectors = [torch.asarray(vector, dtype) for vector in eigenvectors]

    shape = summed_eigenvalues.shape
    if diagonals.shape != shape:
        raise ValueError(
            "output shape from func() does not match input shape: "
            f"{diagonals.shape} vs {shape}"
        )

    def apply(rhs: Array) -> Array:
        if rhs.shape != shape:
            raise ValueError(f"rhs.shape={rhs.shape} does not match shape={shape}")
        if rhs.dtype != dtype:
            raise ValueError(f"rhs.dtype={rhs.dtype} does not match dtype={dtype}")

        # Use tensordot so we have more control over the underlying XLA operations.
        out = rhs
        for vectors in eigenvectors:
            out = torch.tensordot(out, vectors, dims=(0, 0))
        out *= diagonals
        for vectors in eigenvectors:
            out = torch.tensordot(out, vectors, dims=(0, 1))
        return out

    return apply


def _circulant_fft_transform(
    func: Callable[[Array], Array],
    operators: Sequence[Array],
    dtype: torch.dtype,
) -> Callable[[Array], Array]:
    """Fast diagonalization by Fast Fourier Transform."""
    # https://en.wikipedia.org/wiki/Circulant_matrix#Eigenvectors_and_eigenvalues
    eigenvalues = [fft.fft(op[:, 0]) for op in operators]
    summed_eigenvalues = outer_sum(eigenvalues)  # e[m,n,k] = m^2 + n^2 + k^2
    diagonals = torch.asarray(func(summed_eigenvalues))

    shape = tuple(op.shape[0] for op in operators)
    if diagonals.shape != shape:
        raise ValueError(
            "output shape from func() does not match input shape: "
            f"{diagonals.shape} vs {shape}"
        )

    def apply(rhs: Array) -> Array:
        if rhs.shape != shape:
            raise ValueError(f"rhs.shape={rhs.shape} does not match shape={shape}")
        return fft.ifftn(diagonals * fft.fftn(rhs)).to(dtype)

    return apply


def _circulant_rfft_transform(
    func: Callable[[Array], Array],
    operators: Sequence[Array],
    dtype: torch.dtype,
) -> Callable[[Array], Array]:
    """Fast diagonalization by real-valued Fast Fourier Transform."""
    # https://en.wikipedia.org/wiki/Circulant_matrix#Eigenvectors_and_eigenvalues
    if operators[-1].shape[0] % 2:
        raise ValueError(
            'implementation="rfft" currently requires an even size ' "for the last axis"
        )
    # Use `rfft()` only on the last operator so the shape of `diagonals` matches
    # the shape of the output from `rfftn()` without any extra wrangling.
    eigenvalues = [fft.fft(op[:, 0]) for op in operators[:-1]] + [
        fft.rfft(operators[-1][:, 0])
    ]
    summed_eigenvalues = outer_sum(eigenvalues)
    diagonals = torch.asarray(func(summed_eigenvalues))

    if diagonals.shape != summed_eigenvalues.shape:
        raise ValueError(
            "output shape from func() does not match input shape: "
            f"{diagonals.shape} vs {summed_eigenvalues.shape}"
        )

    def apply(rhs: Array) -> Array:
        if rhs.dtype != dtype:
            raise ValueError(f"rhs.dtype={rhs.dtype} does not match dtype={dtype}")
        return fft.irfftn(diagonals * fft.rfftn(rhs)).to(dtype)

    return apply


def pseudoinverse(
    v: Array,
    operators: Sequence[Array],
    dtype: torch.dtype,
    *,
    hermitian: bool = False,
    circulant: bool = False,
    implementation: Optional[str] = None,
    cutoff: Optional[float] = None,
) -> Callable[[Array], Array]:
    """Invert a linear operator written as a sum of operators on each axis.

    Args:
      operators: forward linear operators as matrices, applied along each axis.
        Each of these matrices is diagonalized.
      dtype: dtype of the right-hand-side.
      hermitian: whether or not all linear operator are Hermitian (i.e., symmetric
        in the real valued case).
      circulant: whether or not all linear operators are circulant.
      implementation: how to implement fast diagonalization.
      precision: numerical precision for matrix multplication. Only relevant on
        TPUs.
      cutoff: eigenvalues with absolute value smaller than this number are
        discarded rather than being inverted. By default, uses 10 times floating
        point epsilon.

    Returns:
      A function that computes the pseudo-inverse of the indicated operator.
    """
    if cutoff is None:
        cutoff = 10 * torch.finfo(dtype).eps

    def func(v):
        return torch.where(abs(v) > cutoff, 1 / v, 0)
    
    t = transform(
        func,
        operators,
        dtype,
        hermitian=hermitian,
        circulant=circulant,
        implementation=implementation,
    )

    return t(v)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Modifications copyright (C) 2024 S.Cao
# ported Google's Jax-CFD functional template to PyTorch's tensor ops

from typing import Callable, Dict, Optional, Tuple

import torch
import torch.fft as fft
import torch.nn as nn
from tqdm.auto import tqdm

from . import grids

TQDM_ITERS = 500

Array = torch.Tensor
Grid = grids.Grid


def stable_time_step(
    dx: float = None,
    dt: float = None,
    max_velocity: float = 1.0,
    max_courant_number: float = 0.5,
    viscosity: float = 1e-3,
    implicit_diffusion: bool = True,
    ndim: int = 2,
) -> float:
    """
    Calculate a stable time step satisfying the CFL condition
    for the explicit advection term
    if the diffusion is explicit, the time step is the smaller
    of the advection and diffusion time steps.

    Args:
    max_velocity: maximum velocity.
    max_courant_number: the Courant number used to choose the time step. Smaller
      numbers will lead to more stable simulations. Typically this should be in
      the range [0.5, 1).
    dx: spatial mesh size, can be min(grid.step).
    dt: time step.
    """
    dt_diffusion = dx

    if not implicit_diffusion:
        dt_diffusion = dx**2 / (viscosity * 2 ** (ndim))
    dt_advection = max_courant_number * dx / max_velocity
    dt = dt_advection if dt is None else dt
    return min(dt_diffusion, dt_advection, dt)


class ImplicitExplicitODE(nn.Module):
    """Describes a set of ODEs with implicit & explicit terms.

    The equation is given by:

      $\partial x/ \partial t = explicit_terms(x) + implicit_terms(x)$

    `explicit_terms(x)` includes terms that should use explicit time-stepping and
    `implicit_terms(x)` includes terms that should be modeled implicitly.

    Typically the explicit terms are non-linear and the implicit terms are linear.
    This simplifies solves but isn't strictly necessary.
    """

    def explicit_terms(self, *, x):
        """Evaluates explicit terms in the ODE."""
        raise NotImplementedError

    def implicit_terms(self, *, x):
        """Evaluates implicit terms in the ODE."""
        raise NotImplementedError

    def implicit_solve(
        self,
        *,
        x: Array,
        step_size: float,
    ):
        """Solves `y - step_size * implicit_terms(y) = x` for y."""
        raise NotImplementedError


def backward_forward_euler(
    u: torch.Tensor,
    dt: float,
    equation: ImplicitExplicitODE,
) -> Array:
    """Time stepping via forward and backward Euler methods.

    This method is first order accurate.

    Args:
        equation: equation to solve.
        time_step: time step.

    Returns:
        Function that performs a time step.
    """
    F = equation.explicit_terms
    G_inv = equation.implicit_solve

    g = u + dt * F(u)
    u = G_inv(g, dt)

    return u


def imex_crank_nicolson(
    u: torch.Tensor,
    dt: float,
    equation: ImplicitExplicitODE,
) -> Array:
    """Time stepping via forward and backward Euler methods.

    This method is first order accurate.

    Args:
        equation: equation to solve.
        time_step: time step.

    Returns:
        Function that performs a time step.
    """
    F = equation.explicit_terms
    G = equation.implicit_terms
    G_inv = equation.implicit_solve

    g = u + dt * F(u) + 0.5 * dt * G(u)
    u = G_inv(g, 0.5 * dt)

    return u


def rk2_crank_nicolson(
    u: torch.Tensor,
    dt: float,
    equation: ImplicitExplicitODE,
) -> Array:
    """Time stepping via Crank-Nicolson and 2nd order Runge-Kutta (Heun).

    This method is second order accurate.

    Args:
      equation: equation to solve.
      time_step: time step.

    Returns:
      Function that performs a time step.

    Reference:
      Chandler, G. J. & Kerswell, R. R. Invariant recurrent solutions embedded in
      a turbulent two-dimensional Kolmogorov flow. J. Fluid Mech. 722, 554–595
      (2013). https://doi.org/10.1017/jfm.2013.122 (Section 3)
    """
    F = equation.explicit_terms
    G = equation.implicit_terms
    G_inv = equation.implicit_solve

    g = u + 0.5 * dt * G(u)
    h = F(u)
    u = G_inv(g + dt * h, 0.5 * dt)
    h = 0.5 * (F(u) + h)
    u = G_inv(g + dt * h, 0.5 * dt)
    return u


def low_storage_runge_kutta_crank_nicolson(
    u: torch.Tensor,
    dt: float,
    params: Dict,
    equation: ImplicitExplicitODE,
) -> Array:
    """
    ported from jax functional programming to be tensor2tensor
    Time stepping via "low-storage" Runge-Kutta and Crank-Nicolson steps.

    These scheme are second order accurate for the implicit terms, but potentially
    higher order accurate for the explicit terms. This seems to be a favorable
    tradeoff when the explicit terms dominate, e.g., for modeling turbulent
    fluids.

    Per Canuto: "[these methods] have been widely used for the time-discretization
    in applications of spectral methods."

    Args:
      alphas: alpha coefficients.
      betas: beta coefficients.
      gammas: gamma coefficients.
      equation.F: explicit terms (convection, rhs, drag).
      equation.G: implicit terms (diffusion).
      equation.implicit_solve: implicit solver, when evaluates at an input (B, n, n), outputs (B, n, n).
      dt: time step.

    Input: w^{t_i} (B, n, n)
    Returns: w^{t_{i+1}} (B, n, n)

    Reference:
      Canuto, C., Yousuff Hussaini, M., Quarteroni, A. & Zang, T. A.
      Spectral Methods: Evolution to Complex Geometries and Applications to
      Fluid Dynamics. (Springer Berlin Heidelberg, 2007).
      https://doi.org/10.1007/978-3-540-30728-0 (Appendix D.3)
    """
    dt = dt
    alphas = params["alphas"]
    betas = params["betas"]
    gammas = params["gammas"]
    F = equation.explicit_terms
    G = equation.implicit_terms
    G_inv = equation.implicit_solve

    if len(alphas) - 1 != len(betas) != len(gammas):
        raise ValueError("number of RK coefficients does not match")

    h = 0
    for k in range(len(betas)):
        h = F(u) + betas[k] * h
        mu = 0.5 * dt * (alphas[k + 1] - alphas[k])
        u = G_inv(u + gammas[k] * dt * h + mu * G(u), mu)
    return u


def crank_nicolson_rk4(
    u: Array,
    dt: float,
    equation: ImplicitExplicitODE,
) -> Array:
    """Time stepping via Crank-Nicolson and RK4 ("Carpenter-Kennedy")."""
    params = dict(
        alphas=[
            0,
            0.1496590219993,
            0.3704009573644,
            0.6222557631345,
            0.9582821306748,
            1,
        ],
        betas=[0, -0.4178904745, -1.192151694643, -1.697784692471, -1.514183444257],
        gammas=[
            0.1496590219993,
            0.3792103129999,
            0.8229550293869,
            0.6994504559488,
            0.1530572479681,
        ],
    )
    return low_storage_runge_kutta_crank_nicolson(
        u,
        dt=dt,
        equation=equation,
        params=params,
    )


class NavierStokes2DSpectral(ImplicitExplicitODE):
    """Breaks the Navier-Stokes equation into implicit and explicit parts.

    Implicit parts are the linear terms and explicit parts are the non-linear
    terms.

    Attributes:
      viscosity: strength of the diffusion term
      grid: underlying grid of the process
      smooth: smooth the advection term using the 2/3-rule.
      forcing_fn: forcing function, if None then no forcing is used.
      drag: strength of the drag. Set to zero for no drag.
    """

    def __init__(
        self,
        viscosity: float,
        grid: Grid,
        drag: float = 0.0,
        smooth: bool = True,
        forcing_fn: Optional[Callable] = None,
        solver: Optional[Callable] = crank_nicolson_rk4,
    ):
        super().__init__()
        self.viscosity = viscosity
        self.grid = grid
        self.drag = drag
        self.smooth = smooth
        self.forcing_fn = forcing_fn
        self.solver = solver
        self._initialize()

    def _initialize(self):
        kx, ky = self.grid.rfft_mesh()
        self.register_buffer("kx", kx)
        self.register_buffer("ky", ky)
        laplace = (torch.pi * 2j) ** 2 * (self.kx**2 + self.ky**2)
        self.register_buffer("laplace", laplace)
        filter_ = self.brick_wall_filter_2d(self.grid)
        linear_term = self.viscosity * self.laplace - self.drag
        self.register_buffer("linear_term", linear_term)
        self.register_buffer("filter", filter_)

    @staticmethod
    def brick_wall_filter_2d(grid: Grid):
        """Implements the 2/3 rule."""
        n, _ = grid.shape
        filter_ = torch.zeros((n, n // 2 + 1))
        filter_[: int(2 / 3 * n) // 2, : int(2 / 3 * (n // 2 + 1))] = 1
        filter_[-int(2 / 3 * n) // 2 :, : int(2 / 3 * (n // 2 + 1))] = 1
        return filter_

    @staticmethod
    def spectral_curl_2d(vhat, rfft_mesh):
        r"""
        Computes the 2D curl in the Fourier basis.
        det [d_x d_y \\ u v]
        """
        uhat, vhat = vhat
        kx, ky = rfft_mesh
        return 2j * torch.pi * (vhat * kx - uhat * ky)

    @staticmethod
    def spectral_grad_2d(vhat, rfft_mesh):
        kx, ky = rfft_mesh
        return 2j * torch.pi * kx * vhat, 2j * torch.pi * ky * vhat

    @staticmethod
    def vorticity_to_velocity(
        grid: Grid, w_hat: Array, rfft_mesh: Optional[Tuple[Array, Array]] = None
    ):
        """Constructs a function for converting vorticity to velocity, both in Fourier domain.

        Solves for the stream function and then uses the stream function to compute
        the velocity. This is the standard approach. A quick sketch can be found in
        [1].

        Args:
            grid: the grid underlying the vorticity field.

        Returns:
            A function that takes a vorticity (rfftn) and returns a velocity vector
            field.

        Reference:
            [1] Z. Yin, H.J.H. Clercx, D.C. Montgomery, An easily implemented task-based
            parallel scheme for the Fourier pseudospectral solver applied to 2D
            Navier-Stokes turbulence, Computers & Fluids, Volume 33, Issue 4, 2004,
            Pages 509-520, ISSN 0045-7930,
            https://doi.org/10.1016/j.compfluid.2003.06.003.
        """
        kx, ky = rfft_mesh if rfft_mesh is not None else grid.rfft_mesh()
        two_pi_i = 2 * torch.pi * 1j
        assert kx.shape[-2:] == w_hat.shape[-2:]
        laplace = two_pi_i**2 * (abs(kx) ** 2 + abs(ky) ** 2)
        laplace[..., 0, 0] = 1
        psi_hat = -1 / laplace * w_hat
        u_hat = two_pi_i * ky * psi_hat
        v_hat = -two_pi_i * kx * psi_hat
        return u_hat, v_hat

    def residual(
        self,
        vort_hat: Array,
        vort_t_hat: Array,
    ):
        residual = (
            vort_t_hat
            - self.explicit_terms(vort_hat)
            - self.viscosity * self.implicit_terms(vort_hat)
        )
        return residual

    def _explicit_terms(self, vort_hat):
        vxhat, vyhat = self.vorticity_to_velocity(
            self.grid, vort_hat, (self.kx, self.ky)
        )
        vx, vy = fft.irfft2(vxhat), fft.irfft2(vyhat)

        grad_x_hat = 2j * torch.pi * self.kx * vort_hat
        grad_y_hat = 2j * torch.pi * self.ky * vort_hat
        grad_x, grad_y = fft.irfft2(grad_x_hat), fft.irfft2(grad_y_hat)

        advection = -(grad_x * vx + grad_y * vy)
        advection_hat = fft.rfft2(advection)

        if self.smooth:
            advection_hat *= self.filter

        terms = advection_hat

        if self.forcing_fn is not None:
            if not self.forcing_fn.vorticity:
                fx, fy = self.forcing_fn(self.grid, (vx, vy))
                fx_hat, fy_hat = fft.rfft2(fx.data), fft.rfft2(fy.data)
                terms += self.spectral_curl_2d((fx_hat, fy_hat), (self.kx, self.ky))
            else:
                f = self.forcing_fn(self.grid, vort_hat)
                f_hat = fft.rfft2(f.data)
                terms += f_hat
        return terms

    def explicit_terms(self, vort_hat):
        return self._explicit_terms(vort_hat)

    def implicit_terms(self, vort_hat):
        return self.linear_term * vort_hat

    def implicit_solve(self, vort_hat, dt):
        return 1 / (1 - dt * self.linear_term) * vort_hat

    def get_trajectory(
        self,
        w0: Array,
        dt: float,
        T: float,
        record_every_steps=1,
        pbar=False,
        pbar_desc="",
        require_grad=False,
    ):
        """
        vorticity stacked in the time dimension
        all inputs and outputs are in the frequency domain
        input: w0 (*, n, n)
        output:

        vorticity (*, n_t, kx, ky)
        velocity: tuple of (*, n_t, kx, ky)
        """
        w_all = []
        u_all = []
        v_all = []
        dwdt_all = []
        res_all = []
        w = w0
        time_steps = int(T / dt)
        update_iters = time_steps // TQDM_ITERS
        with tqdm(total=time_steps) as pbar:
            for t in range(time_steps):
                w, dwdt = self.forward(w, dt=dt)
                w.requires_grad_(require_grad)
                dwdt.requires_grad_(require_grad)

                if t % update_iters == 0:
                    pbar.set_description(pbar_desc)
                    pbar.update(update_iters)

                if t % record_every_steps == 0:
                    u, v = self.vorticity_to_velocity(self.grid, w)
                    res = self.residual(w, dwdt)

                    w_, dwdt_, u, v, res = [
                        var.detach().cpu().clone() for var in [w, dwdt, u, v, res]
                    ]

                    w_all.append(w_)
                    u_all.append(u)
                    v_all.append(v)
                    dwdt_all.append(dwdt_)
                    res_all.append(res)

        result = {
            var_name: torch.stack(var, dim=-3)
            for var_name, var in zip(
                ["vorticity", "u", "v", "vort_t", "residual"],
                [w_all, u_all, v_all, dwdt_all, res_all],
            )
        }
        return result

    def step(self, *args, **kwargs):
        return self.forward(*args, **kwargs)

    def forward(self, vort_hat, dt, steps=1):
        """
        vort_hat: (B, kx, ky) or (n_t, kx, ky) or (kx, ky)
        - if rfft2 is used then the shape is (*, kx, ky//2+1)
        - if (n_t, kx, ky), then the time step marches in the time
        dimension in parallel.
        """
        vort_old = vort_hat
        for _ in range(steps):
            vort_hat = self.solver(vort_hat, dt, self)
        dvortdt_hat = 1 / (steps * dt) * (vort_hat - vort_old)
        return vort_hat, dvortdt_hat



from setuptools import setup, find_packages

setup(
  name = 'torch-cfd',
  packages=find_packages(include=['torch_cfd', 'torch_cfd.*']),
  version='0.0.5',
  license='Apache-2.0',
  description = 'PyTorch CFD',
  long_description='PyTorch Computational Fluid Dynamics Library',
  long_description_content_type="text/markdown",
  author = 'Shuhao Cao',
  author_email = 'scao.math@gmail.com',
  url = 'https://github.com/scaomath/torch-cfd',
  keywords = ['pytorch', 'cfd', 'pde', 'spectral'],
  install_requires=[
      'numpy',
      'torch>=2.0.1',
  ],
  classifiers=[
      'Development Status :: 4 - Beta',
      'Intended Audience :: Science/Research',
      'Topic :: Scientific/Engineering :: Mathematics',
      'License :: OSI Approved :: Apache Software License',
      'Programming Language :: Python :: 3.8',
  ],
)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Modifications copyright (C) 2024 S.Cao
# ported Google's Jax-CFD functional template to PyTorch's tensor ops

"""Functions for approximating derivatives.

Finite difference methods operate on GridVariable and return GridArray.
Evaluating finite differences requires boundary conditions, which are defined
for a GridVariable. But the operation of taking a derivative makes the boundary
condition undefined, which is why the return type is GridArray.

For example, if the variable c has the boundary condition c_b = 0, and we take
the derivate dc/dx, then it is unclear what the boundary condition on dc/dx
should be. So programmatically, after taking finite differences and doing
operations, the user has to explicitly assign boundary conditions to the result.

Example:
  c = GridVariable(c_array, c_boundary_condition)
  dcdx = finite_differences.forward_difference(c)  # returns GridArray
  c_new = c + dt * (-velocity * dcdx)  # operations on GridArrays
  c_new = GridVariable(c_new, c_boundary_condition)  # assocaite BCs
"""

import typing
from typing import Optional, Sequence, Tuple
from . import grids
import numpy as np
import torch

Array = torch.Tensor
GridArray = grids.GridArray
GridVariable = grids.GridVariable
GridArrayTensor = grids.GridArrayTensor
GridVariableVector = grids.GridVariableVector


def stencil_sum(*arrays: GridArray) -> GridArray:
    """Sum arrays across a stencil, with an averaged offset."""
    offset = grids.averaged_offset(*arrays)
    # pytype appears to have a bad type signature for sum():
    # Built-in function sum was called with the wrong arguments [wrong-arg-types]
    #          Expected: (iterable: Iterable[complex])
    #   Actually passed: (iterable: Generator[Union[jax.interpreters.xla.DeviceArray, numpy.ndarray], Any, None])
    result = sum(array.data for array in arrays)  # type: ignore
    grid = grids.consistent_grid(*arrays)
    return grids.GridArray(result, offset, grid)


@typing.overload
def forward_difference(u: GridVariable, axis: int) -> GridArray: ...


@typing.overload
def forward_difference(
    u: GridVariable, axis: Optional[Tuple[int, ...]] = ...
) -> Tuple[GridArray, ...]: ...


def forward_difference(u, axis=None):
    """Approximates grads with finite differences in the forward direction."""
    if axis is None:
        axis = range(u.grid.ndim)
    if not isinstance(axis, int):
        return tuple(
            forward_difference(u, a) for a in axis
        )  # pytype: disable=wrong-arg-types  # always-use-return-annotations
    diff = stencil_sum(u.shift(+1, axis), -u.array)
    return diff / u.grid.step[axis]


@typing.overload
def central_difference(u: GridVariable, axis: int) -> GridArray: ...


@typing.overload
def central_difference(
    u: GridVariable, axis: Optional[Tuple[int, ...]]
) -> Tuple[GridArray, ...]: ...


def central_difference(u, axis=None):
    """Approximates grads with central differences."""
    if axis is None:
        axis = range(u.grid.ndim)
    if not isinstance(axis, int):
        return tuple(central_difference(u, a) for a in axis)
    diff = stencil_sum(u.shift(+1, axis), -u.shift(-1, axis))
    return diff / (2 * u.grid.step[axis])


@typing.overload
def backward_difference(u: GridVariable, axis: int) -> GridArray: ...


@typing.overload
def backward_difference(
    u: GridVariable, axis: Optional[Tuple[int, ...]]
) -> Tuple[GridArray, ...]: ...


def backward_difference(u, axis=None):
    """Approximates grads with finite differences in the backward direction."""
    if axis is None:
        axis = range(u.grid.ndim)
    if not isinstance(axis, int):
        return tuple(backward_difference(u, a) for a in axis)
    diff = stencil_sum(u, -u.shift(-1, axis))
    return diff / u.grid.step[axis]


def divergence(v: Sequence[GridVariable]) -> GridArray:
    """Approximates the divergence of `v` using backward differences."""
    grid = grids.consistent_grid(*v)
    if len(v) != grid.ndim:
        raise ValueError(
            "The length of `v` must be equal to `grid.ndim`."
            f"Expected length {grid.ndim}; got {len(v)}."
        )
    differences = [backward_difference(u, axis) for axis, u in enumerate(v)]
    return sum(differences)


def centered_divergence(v: Sequence[GridVariable]) -> GridArray:
    """Approximates the divergence of `v` using centered differences."""
    grid = grids.consistent_grid(*v)
    if len(v) != grid.ndim:
        raise ValueError(
            "The length of `v` must be equal to `grid.ndim`."
            f"Expected length {grid.ndim}; got {len(v)}."
        )
    differences = [central_difference(u, axis) for axis, u in enumerate(v)]
    return sum(differences)


def laplacian_matrix(n: int, step: float) -> Array:
    """
    Create 1D Laplacian operator matrix, with periodic BC.
    modified the scipy.linalg.circulant implementation to native torch
    """
    column = torch.zeros(n)
    column[0] = -2 / step**2
    column[1] = column[-1] = 1 / step**2
    idx = (n - torch.arange(n)[None].T + torch.arange(n)[None]) % n
    return torch.gather(column[None, ...].expand(n, -1), 1, idx)


def _linear_along_axis(c: GridVariable, offset: float, axis: int) -> GridVariable:
    """Linear interpolation of `c` to `offset` along a single specified `axis`."""
    offset_delta = offset - c.offset[axis]

    # If offsets are the same, `c` is unchanged.
    if offset_delta == 0:
        return c

    new_offset = tuple(offset if j == axis else o for j, o in enumerate(c.offset))

    # If offsets differ by an integer, we can just shift `c`.
    if int(offset_delta) == offset_delta:
        return grids.GridVariable(
            array=grids.GridArray(
                data=c.shift(int(offset_delta), axis).data,
                offset=new_offset,
                grid=c.grid,
            ),
            bc=c.bc,
        )

    floor = int(np.floor(offset_delta))
    ceil = int(np.ceil(offset_delta))
    floor_weight = ceil - offset_delta
    ceil_weight = 1.0 - floor_weight
    data = (
        floor_weight * c.shift(floor, axis).data
        + ceil_weight * c.shift(ceil, axis).data
    )
    return grids.GridVariable(array=grids.GridArray(data, new_offset, c.grid), bc=c.bc)


def linear(
    c: GridVariable,
    offset: Tuple[float, ...],
    v: Optional[GridVariableVector] = None,
    dt: Optional[float] = None,
) -> grids.GridVariable:
    """Multi-linear interpolation of `c` to `offset`.

    Args:
      c: quantitity to be interpolated.
      offset: offset to which we will interpolate `c`. Must have the same length
        as `c.offset`.
      v: velocity field. Not used.
      dt: size of the time step. Not used.

    Returns:
      An `GridArray` containing the values of `c` after linear interpolation
      to `offset`. The returned value will have offset equal to `offset`.
    """
    del v, dt  # unused
    if len(offset) != len(c.offset):
        raise ValueError(
            "`c.offset` and `offset` must have the same length;"
            f"got {c.offset} and {offset}."
        )
    interpolated = c
    for a, o in enumerate(offset):
        interpolated = _linear_along_axis(interpolated, offset=o, axis=a)
    return interpolated


@typing.overload
def gradient_tensor(v: GridVariable) -> GridArrayTensor: ...


@typing.overload
def gradient_tensor(v: Sequence[GridVariable]) -> GridArrayTensor: ...


def gradient_tensor(v):
    """Approximates the cell-centered gradient of `v`."""
    if not isinstance(v, GridVariable):
        return GridArrayTensor(np.stack([gradient_tensor(u) for u in v], axis=-1))
    grad = []
    for axis in range(v.grid.ndim):
        offset = v.offset[axis]
        if offset == 0:
            derivative = forward_difference(v, axis)
        elif offset == 1:
            derivative = backward_difference(v, axis)
        elif offset == 0.5:
            v_centered = linear(v, v.grid.cell_center)
            derivative = central_difference(v_centered, axis)
        else:
            raise ValueError(f"expected offset values in {{0, 0.5, 1}}, got {offset}")
        grad.append(derivative)
    return GridArrayTensor(grad)


def curl_2d(v: Sequence[GridVariable]) -> GridArray:
    """Approximates the curl of `v` in 2D using forward differences."""
    if len(v) != 2:
        raise ValueError(f"Length of `v` is not 2: {len(v)}")
    grid = grids.consistent_grid(*v)
    if grid.ndim != 2:
        raise ValueError(f"Grid dimensionality is not 2: {grid.ndim}")
    return forward_difference(v[1], axis=0) - forward_difference(v[0], axis=1)
